/*
 * Copyright (C) 2000 - NoCrew Productions
 *
 * STsndplay is licensed under the GNU General Public License.
 * For further details read the provided 'License' file.
 *
 */

#include <stdio.h>
#include <ocpuis.h>
#include "ym.h"
#include "vid.h"
#include "mfp.h"
#include "vec.h"
#include "mem.h"

#define CPU_EQUALIZER 96
#define MIN_YM 0

static unsigned long lastcount;
static unsigned long ymcnt;
static int audiofd;
static unsigned char *audiobuffer;

CPUbyte gen_get_byte(CPUaddr addr)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_get_byte(addr);
  } else if (dispatch == 0xff82) {
    return vid_get_byte(addr);
  } else if (dispatch == 0xfffa) {
    return mfp_get_byte(addr);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_get_byte(addr);
  } else {
    return mem_get_byte(addr);
  }
}

CPUword gen_get_word(CPUaddr addr)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_get_word(addr);
  } else if (dispatch == 0xff82) {
    return vid_get_word(addr);
  } else if (dispatch == 0xfffa) {
    return mfp_get_word(addr);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_get_word(addr);
  } else {
    return mem_get_word(addr);
  }
}

CPUlong gen_get_long(CPUaddr addr)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_get_long(addr);
  } else if (dispatch == 0xff82) {
    return vid_get_long(addr);
  } else if (dispatch == 0xfffa) {
    return mfp_get_long(addr);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_get_long(addr);
  } else {
    return mem_get_long(addr);
  }
}

void gen_put_byte(CPUaddr addr, CPUbyte value)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_put_byte(addr, value);
  } else if (dispatch == 0xff82) {
    return vid_put_byte(addr, value);
  } else if (dispatch == 0xfffa) {
    return mfp_put_byte(addr, value);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_put_byte(addr, value);
  } else {
    return mem_put_byte(addr, value);
  }
}

void gen_put_word(CPUaddr addr, CPUword value)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_put_word(addr, value);
  } else if (dispatch == 0xff82) {
    return vid_put_word(addr, value);
  } else if (dispatch == 0xfffa) {
    return mfp_put_word(addr, value);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_put_word(addr, value);
  } else {
    return mem_put_word(addr, value);
  }
}

void gen_put_long(CPUaddr addr, CPUlong value)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_put_long(addr, value);
  } else if (dispatch == 0xff82) {
    return vid_put_long(addr, value);
  } else if (dispatch == 0xfffa) {
    return mfp_put_long(addr, value);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_put_long(addr, value);
  } else {
    return mem_put_long(addr, value);
  }
}

void *gen_real_addr(CPUaddr addr)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_real_addr(addr);
  } else if (dispatch == 0xff82) {
    return vid_real_addr(addr);
  } else if (dispatch == 0xfffa) {
    return mfp_real_addr(addr);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_real_addr(addr);
  } else {
    return mem_real_addr(addr);
  }
}

int gen_valid_addr(CPUaddr addr, CPUlong size)
{
  int dispatch;

  dispatch = (addr & 0x00ffff00) >> 8;

  if(dispatch == 0xff88) {
    return ym_valid_addr(addr, size);
  } else if (dispatch == 0xff82) {
    return vid_valid_addr(addr, size);
  } else if (dispatch == 0xfffa) {
    return mfp_valid_addr(addr, size);
  } else if ((dispatch == 0x0000) ||
	     (dispatch == 0x0001) ||
	     (dispatch == 0x0002) ||
	     (dispatch == 0x0003)) {
    return vec_valid_addr(addr, size);
  } else {
    return mem_valid_addr(addr, size);
  }
}

void gen_handle_exception(int nr, CPUaddr oldpc)
{
  return;
}

void gen_handle_event(unsigned long cnt)
{
  signed long tcnt;
  
  tcnt = cnt - lastcount;
  if(tcnt < 0) tcnt += 8012800;
  ymcnt += tcnt;
  lastcount = cnt;

#if 1
  if(!mfp_timerrunning(MFP_TA)) {
    if(mfp_timerenable(MFP_TA)) {
      if((mfp_subcnt(MFP_TA, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TA)) {
	if(ymcnt > MIN_YM) {
	  ym_create_samples(audiofd, audiobuffer, 
			    (1.0*ymcnt)/8.0128);
	  ymcnt = 0;
	}
	mfp_mark(MFP_TA, 1);
#if 0
	fprintf(stderr, "Starting Timer-A at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
#endif
	CPUexception(MFP_TAVEC);
#if 0
	fprintf(stderr, "Ending Timer-A\n");
#endif
	mfp_mark(MFP_TA, 0);
	mfp_resetcnt(MFP_TA);
      }
    }
    if(!mfp_timerrunning(MFP_TB)) {
      if(mfp_timerenable(MFP_TB)) {
	if((mfp_subcnt(MFP_TB, tcnt*CPU_EQUALIZER) < 0) &&
	   mfp_timermasked(MFP_TB)) {
	  if(ymcnt > MIN_YM) {
	    ym_create_samples(audiofd, audiobuffer, 
			      (1.0*ymcnt)/8.0128);
	    ymcnt = 0;
	  }
	  mfp_mark(MFP_TB, 1);
#if 0
	  fprintf(stderr, "Starting Timer-B at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
#endif
	  CPUexception(MFP_TBVEC);
#if 0
	  fprintf(stderr, "Ending Timer-B\n");
#endif
	  mfp_mark(MFP_TB, 0);
	  mfp_resetcnt(MFP_TB);
	}
      }
      if(!mfp_timerrunning(MFP_TC)) {
	if(mfp_timerenable(MFP_TC)) {
	  if((mfp_subcnt(MFP_TC, tcnt*CPU_EQUALIZER) < 0) &&
	     mfp_timermasked(MFP_TC)) {
	    if(ymcnt > MIN_YM) {
	      ym_create_samples(audiofd, audiobuffer, 
				(1.0*ymcnt)/8.0128);
	      ymcnt = 0;
	    }
	    mfp_mark(MFP_TC, 1);
#if 0
	    fprintf(stderr, "Starting Timer-C at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
#endif
	    CPUexception(MFP_TCVEC);
#if 0
	    fprintf(stderr, "Ending Timer-C\n");
#endif
	    mfp_mark(MFP_TC, 0);
	    mfp_resetcnt(MFP_TC);
	  }
	}
	if(!mfp_timerrunning(MFP_TD)) {
	  if(mfp_timerenable(MFP_TD)) {
	    if((mfp_subcnt(MFP_TD, tcnt*CPU_EQUALIZER) < 0) &&
	       mfp_timermasked(MFP_TD)) {
	      if(ymcnt > MIN_YM) {
		ym_create_samples(audiofd, audiobuffer, 
				  (1.0*ymcnt)/8.0128);
		ymcnt = 0;
	      }
	      mfp_mark(MFP_TD, 1);
#if 0
	      fprintf(stderr, "Starting Timer-D at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
#endif
	      CPUexception(MFP_TDVEC);
#if 0
	      fprintf(stderr, "Ending Timer-D\n");
#endif
	      mfp_mark(MFP_TD, 0);
	      mfp_resetcnt(MFP_TD);
	    }
	  }
	}
      }
    }
  }
#else
  if(!mfp_timerrunning(MFP_TA)) {
    if(mfp_timerenable(MFP_TA)) {
      if((mfp_subcnt(MFP_TA, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TA)) {
	ym_create_samples(audiofd, audiobuffer, 
			  (1.0*ymcnt)/8.0128);
	ymcnt = 0;
	mfp_mark(MFP_TA, 1);
	fprintf(stderr, "Starting Timer-A at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
	CPUexception(MFP_TAVEC);
	fprintf(stderr, "Ending Timer-A\n");
	mfp_mark(MFP_TA, 0);
	mfp_resetcnt(MFP_TA);
      }
    }
  }

  if(!mfp_timerrunning(MFP_TB) &&
     !mfp_timerrunning(MFP_TA)) {
    if(mfp_timerenable(MFP_TB)) {
      if((mfp_subcnt(MFP_TB, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TB)) {
	ym_create_samples(audiofd, audiobuffer, 
			  (1.0*ymcnt)/8.0128);
	ymcnt = 0;
	mfp_mark(MFP_TB, 1);
	fprintf(stderr, "Starting Timer-B at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
	CPUexception(MFP_TBVEC);
	fprintf(stderr, "Ending Timer-B\n");
	mfp_mark(MFP_TB, 0);
	mfp_resetcnt(MFP_TB);
      }
    }
  }

  if(!mfp_timerrunning(MFP_TC) &&
     !mfp_timerrunning(MFP_TB) &&
     !mfp_timerrunning(MFP_TA)) {
    if(mfp_timerenable(MFP_TC)) {
      if((mfp_subcnt(MFP_TC, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TC)) {
	ym_create_samples(audiofd, audiobuffer, 
			  (1.0*ymcnt)/8.0128);
	ymcnt = 0;
	mfp_mark(MFP_TC, 1);
	fprintf(stderr, "Starting Timer-C at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
	CPUexception(MFP_TCVEC);
	fprintf(stderr, "Ending Timer-C\n");
	mfp_mark(MFP_TC, 0);
	mfp_resetcnt(MFP_TC);
      }
    }
  }

  if(!mfp_timerrunning(MFP_TD) &&
     !mfp_timerrunning(MFP_TC) &&
     !mfp_timerrunning(MFP_TB) &&
     !mfp_timerrunning(MFP_TA)) {
    if(mfp_timerenable(MFP_TD)) {
      if((mfp_subcnt(MFP_TD, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TD)) {
	ym_create_samples(audiofd, audiobuffer, 
			  (1.0*ymcnt)/8.0128);
	ymcnt = 0;
	mfp_mark(MFP_TD, 1);
	fprintf(stderr, "Starting Timer-D at 0x%08x\n", (int)gen_get_long(MFP_TAVEC*4));
	CPUexception(MFP_TDVEC);
	fprintf(stderr, "Ending Timer-D\n");
	mfp_mark(MFP_TD, 0);
	mfp_resetcnt(MFP_TD);
      }
    }
  }
#endif

#if 0
  if(mfp_timerenable(MFP_TC) && 
     !mfp_timerrunning(MFP_TC)) {
      if((mfp_subcnt(MFP_TC, tcnt*CPU_EQUALIZER) < 0) &&
	 mfp_timermasked(MFP_TC)) {
	ym_create_samples(audiofd, audiobuffer, 
			  (1.0*ymcnt)/8.0128);
	ymcnt = 0;
	mfp_mark(MFP_TC, 1);	
	CPUexception(MFP_TCVEC);
	mfp_mark(MFP_TC, 0);
	mfp_resetcnt(MFP_TC);
      }
  }
#endif

  return;
}

CPU *cpu_init(CPUaddr pc, CPUaddr sp, CPUword sr, int afd)
{
  CPU *cpu;

  cpu = CPUsimple_template(pc, sp, sr);
  
  cpu->get_byte = gen_get_byte;
  cpu->get_word = gen_get_word;
  cpu->get_long = gen_get_long;
  cpu->put_byte = gen_put_byte;
  cpu->put_word = gen_put_word;
  cpu->put_long = gen_put_long;
  cpu->real_addr = gen_real_addr;
  cpu->valid_addr= gen_valid_addr;

  cpu->handle_exception = gen_handle_exception;
  cpu->handle_event = gen_handle_event;

  lastcount = 0;
  ymcnt = 0;
  audiofd = afd;

  return cpu;
}

void cpu_setbuffer(unsigned char *buffer)
{
  audiobuffer = buffer;
}
