/*
 * Copyright (C) 2000 - NoCrew Productions
 *
 * STsndplay is licensed under the GNU General Public License.
 * For further details read the provided 'License' file.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ocpuis.h>
#include <sys/time.h>
#include <getopt.h>
#include "mem.h"
#include "mfp.h"
#include "vec.h"
#include "cpu.h"
#include "load.h"
#include "ym.h"
#include "audio.h"
#include "ice.h"
#include "info.h"
#include "setup.h"

#define OUT_BITS 16
#define MAX_HEADER 200

int parse_options(int, char **, settings *);

void setup_timer(int num, CPUaddr addr, CPUaddr play, double freq)
{
  gen_put_long(addr, 0x48e7fffe); /* MOVEM.L d0-a6,-(sp) */
  gen_put_word(addr+4, 0x4eb9); /* JSR (xxx).L */
  gen_put_long(addr+6, play); /* xxx == play */
  gen_put_long(addr+10, 0x4cdf7fff); /* MOVEM.L (sp)+,d0-a6 */
  gen_put_word(addr+14, 0x4e73); /* RTE */
  switch(num) {
  case MFP_TA:
    gen_put_long(MFP_TAVEC*4, addr);
    break;
  case MFP_TB:
    gen_put_long(MFP_TBVEC*4, addr);
    break;
  case MFP_TC:
    gen_put_long(MFP_TCVEC*4, addr);
    break;
  case MFP_TD:
    gen_put_long(MFP_TDVEC*4, addr);
    break;
  }
  
  mfp_activate(num, freq);
}

int main(int argc, char *argv[])
{
  unsigned char *buffer;
  settings set;
  unsigned char *tmp;
  CPU *cpu;
  int afd;
  int arg;
  int timer;
  char *playfreq;
  int maxtime;

  arg = parse_options(argc, argv, &set);
  
  mem_init(1048576, 0x400);
  vec_init(0x400, 0);
  mfp_init();
  ym_init(set.stereo, OUT_BITS, set.freq);
  afd = audio_init(set.stereo, OUT_BITS, set.freq, set.sout);
  cpu = cpu_init(0x8100, 0x8000, 0x2300, afd);

  memcpy(gen_real_addr(0x800), ice_depack, 395);
  load_file(argv[arg], gen_real_addr(0x10000));

  tmp = gen_real_addr(0x10000);

  CPUinit(cpu);
  CPUset_areg(0, 0x10000);
  CPUsubroutine(0x800);

  if(set.verbose) {
    fprintf(stderr, "Composer: %s\n", info_find("COMM", tmp, MAX_HEADER));
    fprintf(stderr, "Title: %s\n", info_find("TITL", tmp, MAX_HEADER));
    fprintf(stderr, "Ripper: %s\n", info_find("RIPP", tmp, MAX_HEADER));
    fprintf(stderr, "Conversion: %s\n", info_find("CONV", tmp, MAX_HEADER));
    fprintf(stderr, "Songs: %d\n", info_snum(tmp, MAX_HEADER));
  }

  timer = MFP_TC;
  maxtime = 1000000/50;

  if((playfreq = info_find("TA", tmp, MAX_HEADER))) {
    timer = MFP_TA;
    maxtime = 1000000/atoi(playfreq);
  }
  if((playfreq = info_find("TB", tmp, MAX_HEADER))) {
    timer = MFP_TB;
    maxtime = 1000000/atoi(playfreq);
    fprintf(stderr, "Got playfreq: %s\n", playfreq);
  }
  if((playfreq = info_find("TC", tmp, MAX_HEADER))) {
    timer = MFP_TC;
    maxtime = 1000000/atoi(playfreq);
  }
  if((playfreq = info_find("TD", tmp, MAX_HEADER))) {
    timer = MFP_TD;
    maxtime = 1000000/atoi(playfreq);
  }

  if(set.verbose) {
    fprintf(stderr, "Using Timer-%c at %dHz\n", 
	    timer+'A', 1000000/maxtime);
  }

  buffer = (unsigned char *)malloc(20*(set.stereo+1)*
				   (set.freq*maxtime/1000000));

  cpu_setbuffer(buffer);

  CPUset_dreg(0, set.song);
  CPUsubroutine(0x10000);

  setup_timer(timer, 0x600, 0x10000+8, 1000000/maxtime);

  while(1) {
    CPUnop(64);
  }

  return 0;
}

int parse_options(int argc, char *argv[], settings *set)
{
  int arg;

  struct option long_options[] = { 
    { "verbose", no_argument, NULL, 'v' },
    { "quiet", no_argument, NULL, 'q' },
    { "stereo", no_argument, NULL, 's' },
    { "stdout", no_argument, NULL, 'S' },
    { "freq", required_argument, NULL, 'f'},
    { "song", required_argument, NULL, 'o'},
    { "help", no_argument, NULL, 'h' },
    { "version", no_argument, NULL, 'V' },
    { NULL, 0, NULL, 0 } };

  char *short_options = "vqsf:o:hVS";

  /* default settings */

  set->verbose = set->stereo = set->quiet = set->sout = 0;
  set->freq = 44100;
  set->song = 1;

  while((arg=getopt_long(argc, argv, 
			 short_options, 
			 long_options, NULL)) != EOF) {
    switch(arg) {
    case 'v':
      set->verbose = 1;
      break;
    case 's':
      set->stereo = 1;
      break;
    case 'S':
      set->sout = 1;
      break;
    case 'q':
      set->quiet = 1;
      break;
    case 'f':
      set->freq = atoi(optarg);
      break;
    case 'o':
      set->song = atoi(optarg);
      break;
    case 'h':
      fprintf(stderr, "Usage: %s [OPTION(S)] <filename>\n"
	      "\n"
	      "valid options are:\n"
	      "-v, --verbose      print informational messages\n"
	      "-q, --quiet        don't print anything at all\n"
	      "-s, --stereo       stereo output (USELESS!)\n"
	      "-f, --freq FREQ    set output frequency (default: 44100)\n"
	      "-o, --song SONG    choose song (default: 1)\n"
	      "-S, --stdout       output sample to stdout\n"
	      "-h, --help         this text\n"
	      "-V, --version      print version and exit\n"
	      "\n"
	      ,argv[0]);
      exit(1);
      break;
    case 'V':
      fprintf(stderr, "STsndplay version 0.0.1\n");
      exit(1);
      break;
    default:
    }
  }
  return optind;
}

