/*
 * Copyright (C) 2000 - NoCrew Productions
 *
 * STsndplay is licensed under the GNU General Public License.
 * For further details read the provided 'License' file.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ocpuis.h>
#include <math.h>
#include "common.h"
#include "mfp.h"

#define MFP_BASE 2457600
#define MFP_EQUALIZER 313

#define MFP_REGS 24

#define MFP_GPIP  0
#define MFP_AER   1
#define MFP_DDR   2
#define MFP_IERA  3
#define MFP_IERB  4
#define MFP_IPRA  5
#define MFP_IPRB  6
#define MFP_ISRA  7
#define MFP_ISRB  8
#define MFP_IMRA  9
#define MFP_IMRB  10
#define MFP_VR    11
#define MFP_TACR  12
#define MFP_TBCR  13
#define MFP_TCDCR 14
#define MFP_TADR  15
#define MFP_TBDR  16
#define MFP_TCDR  17
#define MFP_TDDR  18
#define MFP_SCR   19
#define MFP_UCR   20
#define MFP_RSR   21
#define MFP_TSR   22
#define MFP_UDR   23

#define MFP_MA 0x20
#define MFP_MB 0x01
#define MFP_MC 0x20
#define MFP_MD 0x10

#define MFP_TIMERS 4

static int mfpreg[MFP_REGS];
static int cnt[MFP_TIMERS];

static int mfpdiv[8] = {-1,4,10,16,50,64,100,200};

int _regnum(CPUaddr addr)
{
  return ((addr & 0x3e)>>1);
}

CPUbyte mfp_get_byte(CPUaddr addr)
{
  return mfpreg[_regnum(addr)];
}

CPUword mfp_get_word(CPUaddr addr)
{
  fprintf(stderr, "Reading word 0x%08x\n", (int)addr);
  return 0;
}

CPUlong mfp_get_long(CPUaddr addr)
{
  fprintf(stderr, "Reading long 0x%08x\n", (int)addr);
  return 0;
}

void mfp_put_byte(CPUaddr addr, CPUbyte value)
{
  /*  fprintf(stderr, "Writing MFP register %d = 0x%02x\n", 
      _regnum(addr), value); */

  /*  fprintf(stderr, "Freq: %g (%d/%d)\n", mfp_getfreq(MFP_TD),
	  mfpdiv[mfpreg[MFP_TCDCR] & 0x7],
	  mfpreg[MFP_TDDR]); */
  
  mfpreg[_regnum(addr)] = value;
}

void mfp_put_word(CPUaddr addr, CPUword value)
{
  fprintf(stderr, "Writing word 0x%08x\n", (int)addr);
}

void mfp_put_long(CPUaddr addr, CPUlong value)
{
  fprintf(stderr, "Writing long 0x%08x\n", (int)addr);
}

void *mfp_real_addr(CPUaddr addr)
{
  return NULL;
}

int mfp_valid_addr(CPUaddr addr, CPUlong size)
{
  return 0;
}

void mfp_init(void)
{
  int i;
  
  for(i=0;i<MFP_REGS;i++) {
    mfpreg[i]=0;
  }
  mfpreg[MFP_GPIP]=0xff;

  for(i=0;i<MFP_TIMERS;i++) {
    cnt[i] = 0;
  }
}

int mfp_timerenable(int num)
{
  switch(num) {
  case MFP_TA:
    if(!(mfpreg[MFP_TACR]&0x7)) return 0;
    if((mfpreg[MFP_TACR]&0x8)) return 0;
    return mfpreg[MFP_IERA]&MFP_MA;
  case MFP_TB:
    if(!(mfpreg[MFP_TBCR]&0x7)) return 0;
    if((mfpreg[MFP_TBCR]&0x8)) return 0;
    return mfpreg[MFP_IERA]&MFP_MB;
  case MFP_TC:
    if(!(mfpreg[MFP_TCDCR]&0x70)) return 0;
    return mfpreg[MFP_IERB]&MFP_MC;
  case MFP_TD:
    if(!(mfpreg[MFP_TCDCR]&0x7)) return 0;
    return mfpreg[MFP_IERB]&MFP_MD;
  }
  return 0;
}

int mfp_timerpending(int num)
{
  switch(num) {
  case MFP_TA:
    return mfpreg[MFP_IPRA]&MFP_MA;
  case MFP_TB:
    return mfpreg[MFP_IPRA]&MFP_MB;
  case MFP_TC:
    return mfpreg[MFP_IPRB]&MFP_MC;
  case MFP_TD:
    return mfpreg[MFP_IPRB]&MFP_MD;
  }
  return 0;
}

int mfp_timerrunning(int num)
{
  switch(num) {
  case MFP_TA:
    return mfpreg[MFP_ISRA]&MFP_MA;
  case MFP_TB:
    return mfpreg[MFP_ISRA]&MFP_MB;
  case MFP_TC:
    return mfpreg[MFP_ISRB]&MFP_MC;
  case MFP_TD:
    return mfpreg[MFP_ISRB]&MFP_MD;
  }
  return 0;
}

int mfp_timermasked(int num)
{
  switch(num) {
  case MFP_TA:
    return mfpreg[MFP_IMRA]&MFP_MA;
  case MFP_TB:
    return mfpreg[MFP_IMRA]&MFP_MB;
  case MFP_TC:
    return mfpreg[MFP_IMRB]&MFP_MC;
  case MFP_TD:
    return mfpreg[MFP_IMRB]&MFP_MD;
  }
  return 0;
}

void mfp_resetcnt(int num)
{
  int tmpdiv;
  int tmpdata;

  switch(num) {
  case MFP_TA:
    tmpdiv = mfpdiv[mfpreg[MFP_TACR] & 0x7];
    tmpdata = mfpreg[MFP_TADR];
    break;
  case MFP_TB:
    tmpdiv = mfpdiv[mfpreg[MFP_TBCR] & 0x7];
    tmpdata = mfpreg[MFP_TBDR];
    break;
  case MFP_TC:
    tmpdiv = mfpdiv[(mfpreg[MFP_TCDCR] & 0x70)>>4];
    tmpdata = mfpreg[MFP_TCDR];
    break;
  case MFP_TD:
    tmpdiv = mfpdiv[mfpreg[MFP_TCDCR] & 0x7];
    tmpdata = mfpreg[MFP_TDDR];
    break;
  }
  
  if(tmpdiv == -1) return;
  if(tmpdata == 0) return;

  mfp_setcnt(num, tmpdiv * tmpdata * MFP_EQUALIZER);
}

int mfp_getcnt(int num)
{
  return cnt[num];
}

int mfp_subcnt(int num, int value)
{
  cnt[num] -= value;
  return cnt[num];
}

int mfp_addcnt(int num, int value)
{
  cnt[num] += value;
  return cnt[num];
}

int mfp_setcnt(int num, int value)
{
  cnt[num] = value;
  return cnt[num];
}

double mfp_getfreq(int num)
{
  int tmpdiv;
  int tmpdata;

  switch(num) {
  case MFP_TA:
    tmpdiv = mfpdiv[mfpreg[MFP_TACR] & 0x7];
    tmpdata = mfpreg[MFP_TADR];
    break;
  case MFP_TB:
    tmpdiv = mfpdiv[mfpreg[MFP_TBCR] & 0x7];
    tmpdata = mfpreg[MFP_TBDR];
    break;
  case MFP_TC:
    tmpdiv = mfpdiv[(mfpreg[MFP_TCDCR] & 0x70)>>4];
    tmpdata = mfpreg[MFP_TCDR];
    break;
  case MFP_TD:
    tmpdiv = mfpdiv[mfpreg[MFP_TCDCR] & 0x7];
    tmpdata = mfpreg[MFP_TDDR];
    break;
  }

  if(tmpdiv == -1) return 0;
  if(tmpdata == 0) return 0;
  
  return (((double)(MFP_BASE))/((double)(tmpdiv*tmpdata)));
}

void mfp_activate(int num, double freq)
{
  int pdiv,data;
  double diff,sdiff=1000000;
  int bdiv,bdata;

  for(pdiv=0;pdiv<8;pdiv++) {
    for(data=1;data<256;data++) {
      diff = fabs(freq-(((double)MFP_BASE)/((double)(mfpdiv[pdiv]*data))));
      if(diff < sdiff) {
	bdiv = pdiv;
	bdata = data;
	sdiff = diff;
      }
    }
  }
  
  switch(num) {
  case MFP_TA:
    mfpreg[MFP_TADR] = bdata;
    mfpreg[MFP_TACR] = bdiv;
    mfpreg[MFP_IERA] = (mfpreg[MFP_IERA]&(~MFP_MA))|(MFP_MA);
    mfpreg[MFP_IMRA] = (mfpreg[MFP_IMRA]&(~MFP_MA))|(MFP_MA);
    break;
  case MFP_TB:
    mfpreg[MFP_TBDR] = bdata;
    mfpreg[MFP_TBCR] = bdiv;
    mfpreg[MFP_IERA] = (mfpreg[MFP_IERA]&(~MFP_MB))|(MFP_MB);
    mfpreg[MFP_IMRA] = (mfpreg[MFP_IMRA]&(~MFP_MB))|(MFP_MB);
    break;
  case MFP_TC:
    mfpreg[MFP_TCDR] = bdata;
    mfpreg[MFP_TCDCR] = (mfpreg[MFP_TCDCR]&0x8f)|(bdiv<<4);
    mfpreg[MFP_IERB] = (mfpreg[MFP_IERB]&(~MFP_MC))|(MFP_MC);
    mfpreg[MFP_IMRB] = (mfpreg[MFP_IMRB]&(~MFP_MC))|(MFP_MC);
    break;
  case MFP_TD:
    mfpreg[MFP_TDDR] = bdata;
    mfpreg[MFP_TCDCR] = (mfpreg[MFP_TCDCR]&0xf8)|(bdiv);
    mfpreg[MFP_IERB] = (mfpreg[MFP_IERB]&(~MFP_MD))|(MFP_MD);
    mfpreg[MFP_IMRB] = (mfpreg[MFP_IMRB]&(~MFP_MD))|(MFP_MD);
    break;
  }
  
  mfp_resetcnt(num);
}

void mfp_mark(int num, int active)
{
  int sw;

  sw = active?-1:0;
  switch(num) {
  case MFP_TA:
    mfpreg[MFP_ISRA] = (mfpreg[MFP_ISRA]&(~MFP_MA))|(sw&MFP_MA);
    break;
  case MFP_TB:
    mfpreg[MFP_ISRA] = (mfpreg[MFP_ISRA]&(~MFP_MB))|(sw&MFP_MB);
    break;
  case MFP_TC:
    mfpreg[MFP_ISRB] = (mfpreg[MFP_ISRB]&(~MFP_MC))|(sw&MFP_MC);
    break;
  case MFP_TD:
    mfpreg[MFP_ISRB] = (mfpreg[MFP_ISRB]&(~MFP_MD))|(sw&MFP_MD);
    break;
  }
}
