/*
 * Copyright (C) 2000 - NoCrew Productions
 *
 * STsndplay is licensed under the GNU General Public License.
 * For further details read the provided 'License' file.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ocpuis.h>
#include <psg.h>
#include "common.h"

psg *ym;
unsigned char ymregs[16];
int ymchg[16];
int regsel;

#undef DEBUG

CPUbyte ym_get_byte(CPUaddr addr)
{
  CPUaddr inaddr;

#ifdef DEBUG
  fprintf(stderr, "Get byte: $%lx\n", addr);
#endif
  
  inaddr = addr&0xffffff;

  if(inaddr == 0xff8802) {
    return ymregs[regsel];
  }
  if(inaddr == 0xff8800) {
    return ymregs[regsel];
  }
  return 0;
}

CPUword ym_get_word(CPUaddr addr)
{
#ifdef DEBUG
  fprintf(stderr, "Get word: $%lx\n", addr);
#endif
  
  return 0;
}

CPUlong ym_get_long(CPUaddr addr)
{
#ifdef DEBUG
  fprintf(stderr, "Get long: $%lx\n", addr);
#endif
  
  return 0;
}

void ym_put_byte(CPUaddr addr, CPUbyte value)
{
  CPUaddr inaddr;

#ifdef DEBUG
  fprintf(stderr, "Put byte: $%lx\n", addr);
#endif
  
  inaddr = addr&0xffffff;

  if(inaddr == 0xff8800) {
    regsel = value;
    return;
  }
  if(inaddr == 0xff8802) {
    ymregs[regsel] = value;
    ymchg[regsel] = 1;
    return;
  }
}

void ym_put_word(CPUaddr addr, CPUword value)
{
#ifdef DEBUG
  fprintf(stderr, "Put word: $%lx\n", addr);
#endif
  

}

void ym_put_long(CPUaddr addr, CPUlong value)
{
#if 0
  fprintf(stderr, "Put long: $%lx / %08lx\n", addr, value);
#endif

  addr = addr & 0xffffff;
  
  if((addr >= 0xff8800) && (addr < 0xff8900)) {
    regsel = (value & 0xff000000) >> 24;
    ymregs[regsel] = (value & 0xff00) >> 8;
    ymchg[regsel] = 1;
    return;
  }
}

void *ym_real_addr(CPUaddr addr)
{
  return NULL;
}

int ym_valid_addr(CPUaddr addr, CPUlong size)
{
  return 0;
}

void ym_init(int mode, int bits, int freq)
{
  ym = psg_init(mode, bits, freq);
}

void ym_create_samples(int afd, void *buffer, int time)
{
  int i;
  int size;

  size = ((ym->output.bits/8) *
	  (ym->output.mode+1) *
	  ((ym->output.freq * time)/1000000));

  for(i=0; i<14; i++) {
    if(ymchg[i])
      ym->regs[i] = ymregs[i];
#if 0
    if(i == 8) ym->regs[i] >>= 1;
    if(i == 9) ym->regs[i] >>= 1;
    if(i == 10) ym->regs[i] >>= 1;
#endif
    ymchg[i] = 0;
  }

  psg_create_samples(buffer, ym, time);
  write(afd, buffer, size);
}

